@@ -14,31 +14,37 @@ import traceback
 from itertools import chain
 from weakref import ref as weakref
 from datetime import datetime
-from logbook import helpers
-from logbook.concurrency import thread_get_name, thread_get_ident, greenlet_get_ident
+from logbook.concurrency import (
+    thread_get_name, thread_get_ident, greenlet_get_ident)
 
-from logbook.helpers import to_safe_json, parse_iso8601, cached_property, \
-     PY2, u, string_types, iteritems, integer_types
+from logbook.helpers import (
+    to_safe_json, parse_iso8601, cached_property, PY2, u, string_types,
+    iteritems, integer_types, xrange)
 try:
-    from logbook._speedups import group_reflected_property, \
-         ContextStackManager, StackedObject
+    from logbook._speedups import (
+        group_reflected_property, ContextStackManager, StackedObject)
 except ImportError:
-    from logbook._fallback import group_reflected_property, \
-         ContextStackManager, StackedObject
+    from logbook._fallback import (
+        group_reflected_property, ContextStackManager, StackedObject)
 
 _datetime_factory = datetime.utcnow
+
+
 def set_datetime_format(datetime_format):
     """
     Set the format for the datetime objects created, which are then
     made available as the :py:attr:`LogRecord.time` attribute of
     :py:class:`LogRecord` instances.
 
-    :param datetime_format: Indicates how to generate datetime objects.  Possible values are:
+    :param datetime_format: Indicates how to generate datetime objects.
+    Possible values are:
 
          "utc"
-             :py:attr:`LogRecord.time` will be a datetime in UTC time zone (but not time zone aware)
+             :py:attr:`LogRecord.time` will be a datetime in UTC time zone
+             (but not time zone aware)
          "local"
-             :py:attr:`LogRecord.time` will be a datetime in local time zone (but not time zone aware)
+             :py:attr:`LogRecord.time` will be a datetime in local time zone
+             (but not time zone aware)
 
     This function defaults to creating datetime objects in UTC time,
     using `datetime.utcnow()
@@ -66,15 +72,17 @@ def set_datetime_format(datetime_format):
     elif datetime_format == "local":
         _datetime_factory = datetime.now
     else:
-        raise ValueError("Invalid value %r.  Valid values are 'utc' and 'local'." % (datetime_format,))
+        raise ValueError("Invalid value %r.  Valid values are 'utc' and "
+                         "'local'." % (datetime_format,))
 
 # make sure to sync these up with _speedups.pyx
-CRITICAL = 6
-ERROR = 5
-WARNING = 4
-NOTICE = 3
-INFO = 2
-DEBUG = 1
+CRITICAL = 15
+ERROR = 14
+WARNING = 13
+NOTICE = 12
+INFO = 11
+DEBUG = 10
+TRACE = 9
 NOTSET = 0
 
 _level_names = {
@@ -84,6 +92,7 @@ _level_names = {
     NOTICE:     'NOTICE',
     INFO:       'INFO',
     DEBUG:      'DEBUG',
+    TRACE:      'TRACE',
     NOTSET:     'NOTSET'
 }
 _reverse_level_names = dict((v, k) for (k, v) in iteritems(_level_names))
@@ -343,8 +352,8 @@ class LogRecord(object):
     """
     _pullable_information = frozenset((
         'func_name', 'module', 'filename', 'lineno', 'process_name', 'thread',
-        'thread_name', 'greenlet', 'formatted_exception', 'message', 'exception_name',
-        'exception_message'
+        'thread_name', 'greenlet', 'formatted_exception', 'message',
+        'exception_name', 'exception_message'
     ))
     _noned_on_close = frozenset(('exc_info', 'frame', 'calling_frame'))
 
@@ -369,7 +378,8 @@ class LogRecord(object):
     information_pulled = False
 
     def __init__(self, channel, level, msg, args=None, kwargs=None,
-                 exc_info=None, extra=None, frame=None, dispatcher=None, frame_correction=0):
+                 exc_info=None, extra=None, frame=None, dispatcher=None,
+                 frame_correction=0):
         #: the name of the logger that created it or any other textual
         #: channel description.  This is a descriptive name and can be
         #: used for filtering.
@@ -385,8 +395,12 @@ class LogRecord(object):
         #: optional exception information.  If set, this is a tuple in the
         #: form ``(exc_type, exc_value, tb)`` as returned by
         #: :func:`sys.exc_info`.
-        #: This parameter can also be ``True``, which would cause the exception info tuple
-        #: to be fetched for you.
+        #: This parameter can also be ``True``, which would cause the exception
+        #: info tuple to be fetched for you.
+        if not exc_info:
+            # this is a special case where exc_info=False can be passed in
+            # theory, and it should be the same as exc_info=None
+            exc_info = None
         self.exc_info = exc_info
         #: optional extra information as dictionary.  This is the place
         #: where custom log processors can attach custom context sensitive
@@ -495,6 +509,12 @@ class LogRecord(object):
         self.extra = ExtraDict(self.extra)
         return self
 
+    def _format_message(self, msg, *args, **kwargs):
+        """Called if the record's message needs to be formatted.
+        Subclasses can implement their own formatting.
+        """
+        return msg.format(*args, **kwargs)
+
     @cached_property
     def message(self):
         """The formatted message."""
@@ -502,14 +522,17 @@ class LogRecord(object):
             return self.msg
         try:
             try:
-                return self.msg.format(*self.args, **self.kwargs)
+                return self._format_message(self.msg, *self.args,
+                                            **self.kwargs)
             except UnicodeDecodeError:
                 # Assume an unicode message but mixed-up args
                 msg = self.msg.encode('utf-8', 'replace')
-                return msg.format(*self.args, **self.kwargs)
+                return self._format_message(msg, *self.args, **self.kwargs)
             except (UnicodeEncodeError, AttributeError):
-                # we catch AttributeError since if msg is bytes, it won't have the 'format' method
-                if sys.exc_info()[0] is AttributeError and (PY2 or not isinstance(self.msg, bytes)):
+                # we catch AttributeError since if msg is bytes,
+                # it won't have the 'format' method
+                if (sys.exc_info()[0] is AttributeError
+                        and (PY2 or not isinstance(self.msg, bytes))):
                     # this is not the case we thought it is...
                     raise
                 # Assume encoded message with unicode args.
@@ -517,7 +540,7 @@ class LogRecord(object):
                 # but this codepath is unlikely (if the message is a constant
                 # string in the caller's source file)
                 msg = self.msg.decode('utf-8', 'replace')
-                return msg.format(*self.args, **self.kwargs)
+                return self._format_message(msg, *self.args, **self.kwargs)
 
         except Exception:
             # this obviously will not give a proper error message if the
@@ -526,9 +549,9 @@ class LogRecord(object):
             # that.
             e = sys.exc_info()[1]
             errormsg = ('Could not format message with provided '
-                       'arguments: {err}\n  msg={msg!r}\n  '
-                       'args={args!r} \n  kwargs={kwargs!r}.\n'
-                       'Happened in file {file}, line {lineno}').format(
+                        'arguments: {err}\n  msg={msg!r}\n  '
+                        'args={args!r} \n  kwargs={kwargs!r}.\n'
+                        'Happened in file {file}, line {lineno}').format(
                 err=e, msg=self.msg, args=self.args,
                 kwargs=self.kwargs, file=self.filename,
                 lineno=self.lineno
@@ -549,7 +572,10 @@ class LogRecord(object):
         while frm is not None and frm.f_globals is globs:
             frm = frm.f_back
 
-        for _ in helpers.xrange(self.frame_correction):
+        for _ in xrange(self.frame_correction):
+            if frm is None:
+                break
+
             frm = frm.f_back
 
         return frm
@@ -598,8 +624,8 @@ class LogRecord(object):
     @cached_property
     def greenlet(self):
         """The ident of the greenlet.  This is evaluated late and means that
-        if the log record is passed to another greenlet, :meth:`pull_information`
-        was called in the old greenlet.
+        if the log record is passed to another greenlet,
+        :meth:`pull_information` was called in the old greenlet.
         """
         return greenlet_get_ident()
 
@@ -638,7 +664,7 @@ class LogRecord(object):
         """The formatted exception which caused this record to be created
         in case there was any.
         """
-        if self.exc_info is not None:
+        if self.exc_info is not None and self.exc_info != (None, None, None):
             rv = ''.join(traceback.format_exception(*self.exc_info))
             if PY2:
                 rv = rv.decode('utf-8', 'replace')
@@ -693,6 +719,13 @@ class LoggerMixin(object):
     #: created.
     level_name = level_name_property()
 
+    def trace(self, *args, **kwargs):
+        """Logs a :class:`~logbook.LogRecord` with the level set
+        to :data:`~logbook.TRACE`.
+        """
+        if not self.disabled and TRACE >= self.level:
+            self._log(TRACE, args, kwargs)
+
     def debug(self, *args, **kwargs):
         """Logs a :class:`~logbook.LogRecord` with the level set
         to :data:`~logbook.DEBUG`.
@@ -878,7 +911,8 @@ class RecordDispatcher(object):
                 continue
 
             # first case of blackhole (without filter).
-            # this should discard all further processing and we don't have to heavy_init to know that...
+            # this should discard all further processing and
+            # we don't have to heavy_init to know that...
             if handler.filter is None and handler.blackhole:
                 break
 
@@ -891,14 +925,13 @@ class RecordDispatcher(object):
                 self.process_record(record)
                 record_initialized = True
 
-
             # a filter can still veto the handling of the record.  This
             # however is already operating on an initialized and processed
             # record.  The impact is that filters are slower than the
             # handler's should_handle function in case there is no default
             # handler that would handle the record (delayed init).
-            if handler.filter is not None \
-               and not handler.filter(record, handler):
+            if (handler.filter is not None
+                    and not handler.filter(record, handler)):
                 continue
 
             # We might have a filter, so now that we know we *should* handle
@@ -906,7 +939,6 @@ class RecordDispatcher(object):
             if handler.blackhole:
                 break
 
-
             # handle the record.  If the record was handled and
             # the record is not bubbling we can abort now.
             if handler.handle(record) and not handler.bubble:
